
/* $Id: cpd_files.c,v 3.16 1999/11/07 22:20:11 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* File: cpd_files.c -- routines that open/close/read/write/manipulate
	checkpoint files */


/* System Header Files */

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef WIN32
#include <dirent.h>
#include <unistd.h>
#else
#include <io.h>
#endif

extern int alphasort();


/* STV Header Files */

#include "stv.h"
#include "stvmac.h"
#include "cpd_proto.h"
#include "stv_glob.h"


int
cpd_open(path,lsize)
char *path;
long *lsize;
{
int fd;
	*lsize = 0L;
	if ( path == (char *) NULL)
	{
		STV_ERROR( "cpd_open", stvStatusBadParam,
				"Null path name." );
		return( stvStatusBadParam );
	}
	fd = open(path, O_RDWR | O_CREAT, 0644 );
	if ( fd < 0)
	{
		sprintf( STVERRTXT, "Couldn't Open/Create File %s.", path );
		STV_ERROR( "cpd_open", stvStatusSysFail, STVERRTXT );
		return( stvStatusSysFail );
	}
	/* inquire on the length of the file */
	lseek(fd, (off_t) 0, SEEK_END);
	*lsize = lseek(fd, (off_t) 0, SEEK_CUR);
	lseek(fd, (off_t) 0, SEEK_SET);	/* return to the beginning */
	return (fd);
}


/* --------------------------------------------------- */
/* structure of a checkpoint file is as follows:
	HEADER
	------
	ntasks-sized array of offsets -- indicates where node X can find
	its checkpoint data in the file
*/

/*------ write cpd header data ---------- */
int
cpd_write_hdr(fd,header,iflag)
int fd;
CPD_HEADERPTR header;
int iflag;
{
long nwrt;
off_t *offsets;
int i;
	lseek(fd, (off_t) 0 , SEEK_SET);
	write(fd, header, sizeof(CPD_HEADER));
	nwrt = lseek(fd, (off_t) 0, SEEK_CUR);
	if (nwrt != (long) sizeof(CPD_HEADER))
	{
		STV_ERROR( "cpd_write_hdr", stvStatusSysFail,
				"Header Not Written." );
		return( stvStatusSysFail );
	}
	/* allocate and fill the offsets block in the file */
	if (iflag > 0 && header->ntasks > 0)
	{
		if ( (offsets = (off_t *) MALLOC( (STV_SIZE_T) header->ntasks
				* sizeof(off_t) )) != (off_t *) NULL)
		{
			for (i = 0; i < header->ntasks; i++)
				offsets[i] = (off_t) -1;
			nwrt = write(fd, offsets, header->ntasks*sizeof(off_t));
			FREE (offsets);
			if (nwrt != header->ntasks*sizeof(off_t))
			{
				STV_ERROR( "cpd_write_hdr", stvStatusSysFail,
						"Offsets Not Written." );
				return( stvStatusSysFail );
			}
		}
	}
	return (sizeof(CPD_HEADER) + header->ntasks*sizeof(off_t));
}


/* ------- cpd_node_data_buf -------- */
/*
	read the file header to determine where node data should be
	written/read from in the. If intending to write, the data should
	be written immediatly after the call to cpd_node_data_buf.
	returns stvTrue if data has already been written to the file.
			stvFalse if it has not.
*/
int
cpd_node_data_buf(fd,nodeid)
int fd;
int nodeid;
{
off_t datarec;
off_t loc;
	lseek(fd,(off_t) (sizeof(CPD_HEADER) + nodeid*sizeof(off_t)), SEEK_SET);
	read(fd,(char *) &datarec,sizeof(off_t));
	if ( datarec == (off_t) -1)	/* node not written, put at end of file */
	{
		lseek(fd, (off_t) 0, SEEK_END);
		loc = lseek(fd, (off_t) 0, SEEK_CUR);
		lseek(fd,(off_t) (sizeof(CPD_HEADER) + nodeid*sizeof(off_t)),SEEK_SET);
		write(fd,(char *)&loc, sizeof (loc));
		lseek(fd,(off_t) 0, SEEK_END);
		return (stvFalse); /* data has not been written */
	}
	else
	{
		lseek(fd,datarec,SEEK_SET);
		return(stvTrue);	/* data has been written */
	}
}


int
cpd_read_header(fd,header)
int fd;
CPD_HEADERPTR header;
{
long rval;
	rval = lseek (fd, (off_t) 0, SEEK_SET);
	if (rval == -1L)
		return(-1);
	return (read(fd, (char *) header, sizeof(CPD_HEADER)));
}


CPD_PARTICLE_COORDS cpd_create_particle_coords()
{
	CPD_PARTICLE_COORDS tmp;

	int i;

	tmp = (CPD_PARTICLE_COORDS) MALLOC(
		sizeof( struct cpd_particle_coords ) );
	STV_MEMCHECK( tmp, "Checkpoint Particle Coords Structure",
		return( (CPD_PARTICLE_COORDS) NULL ) );

	for ( i=0 ; i < STV_MAX_DIM ; i++ )
		tmp->coords[i] = -1;

	tmp->ptid = (STV_PARTICLE_ID) NULL;

	tmp->next = (CPD_PARTICLE_COORDS) NULL;

	return( tmp );
}


void
cpd_free_particle_coords( ptr )
CPD_PARTICLE_COORDS *ptr;
{
	CPD_PARTICLE_COORDS C;

	int i;

	if ( ptr == NULL || *ptr == NULL )
		return;

	C = *ptr;

	for ( i=0 ; i < STV_MAX_DIM ; i++ )
		C->coords[i] = -1;

	C->ptid = (STV_PARTICLE_ID) NULL;

	C->next = (CPD_PARTICLE_COORDS) NULL;

	FREE( C );

	*ptr = (CPD_PARTICLE_COORDS) NULL;
}


CPD_PARTICLE_DATA cpd_create_particle_data()
{
	CPD_PARTICLE_DATA tmp;

	int i;

	tmp = (CPD_PARTICLE_DATA) MALLOC(
		sizeof( struct cpd_particle_data ) );
	STV_MEMCHECK( tmp, "Checkpoint Particle Data Structure",
		return( (CPD_PARTICLE_DATA) NULL ) );

	tmp->value = (STV_VALUE) NULL;

	tmp->next = (CPD_PARTICLE_DATA) NULL;

	return( tmp );
}


void
cpd_free_particle_data( ptr )
CPD_PARTICLE_DATA *ptr;
{
	CPD_PARTICLE_DATA PD;

	int i;

	if ( ptr == NULL || *ptr == NULL )
		return;

	PD = *ptr;

	PD->value = (STV_VALUE) NULL;

	PD->next = (CPD_PARTICLE_DATA) NULL;

	FREE( PD );

	*ptr = (CPD_PARTICLE_DATA) NULL;
}


CPD_DATAPTR cpd_create_data()
{
	CPD_DATAPTR tmp;

	tmp = (CPD_DATAPTR) MALLOC( sizeof( struct cpd_data ) );
	STV_MEMCHECK( tmp, "Checkpoint Data Structure",
		return( (CPD_DATAPTR) NULL ) );

	tmp->param = (STV_PARAMPTR) NULL;

	tmp->particle = (STV_PARTICLEPTR) NULL;

	tmp->pcoords = (CPD_PARTICLE_COORDS) NULL;

	tmp->field = (STV_FIELDPTR) NULL;

	tmp->pdata = (CPD_PARTICLE_DATA) NULL;

	tmp->next = (CPD_DATAPTR) NULL;

	return( tmp );
}


void
cpd_free_data( ptr )
CPD_DATAPTR *ptr;
{
	CPD_DATAPTR D;

	if ( ptr == NULL || *ptr == NULL )
		return;

	D = *ptr;

	D->param = (STV_PARAMPTR) NULL;

	D->particle = (STV_PARTICLEPTR) NULL;

	D->pcoords = (CPD_PARTICLE_COORDS) NULL;

	D->field = (STV_FIELDPTR) NULL;

	D->pdata = (CPD_PARTICLE_DATA) NULL;

	D->next = (CPD_DATAPTR) NULL;

	FREE( D );

	*ptr = (CPD_DATAPTR) NULL;
}


int
cpd_write_str( fd, str )
int fd;
char *str;
{
	int len;
	int cc;

	if ( str == NULL )
		len = 0;

	else
		len = strlen( str ) + 1;

	cc = write( fd, &len, sizeof( int ) );

	if ( cc < 0 )
	{
		perror( "cpd_write_str - len" );

		return( stvFalse );
	}

	cc = write( fd, str, (STV_SIZE_T) len * sizeof( char ) );

	if ( cc < 0 )
	{
		perror( "cpd_write_str - str" );

		return( stvFalse );
	}

	return( stvTrue );
}


char *cpd_read_str( fd )
int fd;
{
	char *str;

	int len;
	int cc;

	cc = read( fd, &len, sizeof( int ) );

	if ( cc < 0 )
	{
		perror( "cpd_read_str - len" );

		return( (char *) NULL );
	}

	str = (char *) MALLOC( (STV_SIZE_T) len * sizeof( char ) );
	STV_MEMCHECK( str, "CPD Read String", return( (char *) NULL ) );

	cc = read( fd, str, (STV_SIZE_T) len * sizeof( char ) );

	if ( cc < 0 )
	{
		perror( "cpd_read_str - str" );

		FREE( str );

		return( (char *) NULL );
	}

	return( str );
}


int
cpd_get_fnames(dir, namelist)
char * dir;
struct dirent ***namelist;
{
	char cpd_dir[stvMaxNameLen];
	int i, n;
	struct stat statbuf;

	if (!dir)
		strcpy(cpd_dir,".");	/* current directory */
	else
		strcpy(cpd_dir,dir);

	if (stat(cpd_dir,&statbuf) < 0)
	{
		perror("cpd_list_chkpts(): stat failed");
		return -1;
	}
	n = scandir(cpd_dir, namelist, 0, alphasort);
	if (n < 0 )
		perror("scandir");

	return n;
}


void
cpd_print_fnames(namelist, n)
struct dirent **namelist;
int n;
{
	int i;

	if (n)
		printf( "Listing Checkpoint files:\n" );
	else
		printf( "No Checkpoint files.\n" );

	for (i = 0; i < n; i ++)
		printf("\t%s\n", namelist[i]->d_name);
}


int
cpd_pack_file_list(dir, bufid)
char *dir;
int bufid;
{
struct dirent **namelist;
int i, fd, nfiles;
CPD_HEADER header;
int bogus = stvFalse;
int valid = stvTrue;
long nbytes;
struct stat statbuf;
char fname[stvMaxNameLen];

	nfiles = cpd_get_fnames(dir, &namelist);

	stv_pack_int(bufid, &nfiles, 1, 1);

	for (i = 0; i < nfiles; i ++)
	{
		sprintf(fname,"%s/%s",dir,namelist[i]->d_name);

		if (stat(fname,&statbuf) < 0)
		{
			perror("pack_file_list(): stat failed");
			stv_pack_int(bufid, &bogus, 1, 1);
			continue;
		}

		if (S_ISDIR(statbuf.st_mode))
		{
			stv_pack_int(bufid, &bogus, 1, 1);
			continue;
		}

		if ( (fd = cpd_open(fname, &nbytes)) < 0)
		{
			printf( "cpd_pack_file_list(): Could not open %s.\n",
					namelist[i]->d_name );
			stv_pack_int(bufid, &bogus, 1, 1);
		}
		else
		{
			if (nbytes < sizeof(header))
				stv_pack_int(bufid, &bogus, 1, 1);
			else
			{
				cpd_read_header(fd, &header);
				if (header.complete)
				{
					stv_pack_int(bufid, &valid, 1, 1);
					stv_pack_str(bufid,namelist[i]->d_name);
					stv_pack_str(bufid, header.name);
					stv_pack_str(bufid, header.aout);
					stv_pack_str(bufid, header.arch);
					stv_pack_int(bufid, &header.ntasks,1,1);
					stv_pack_int(bufid, &header.epoch,1,1);
				}
				else
				{
					stv_pack_int(bufid, &bogus, 1, 1);
				}

			}
			close(fd);

		}

	}
	if (namelist)
		FREE(namelist);		/* malloc'ed by scandir */

	return stvStatusOk;
}


int
cpd_protoCpdFileList( msg )
STV_BUFPTR msg;
{
	char tststr[ stvMaxNameLen ];

	int cc, i;
	int nfiles;
	int valid;

	stv_unpack_int( msg->bufid, &nfiles, 1, 1 );

	if ( nfiles <= 0 )
		printf( "No files to list:\n" );

	else
	{
		for ( i=0 ; i < nfiles ; i++ )
		{
			stv_unpack_int( msg->bufid, &valid, 1, 1 );

			if ( valid )
			{
				stv_unpack_str( msg->bufid, tststr );
				printf( "FILE: %s\n", tststr );

				stv_unpack_str( msg->bufid, tststr );
				printf( "\tapplication %s\n", tststr );

				stv_unpack_str( msg->bufid, tststr );
				printf( "\texecutable name: %s\n", tststr );

				stv_unpack_str( msg->bufid, tststr );
				printf( "\tarchitecture: %s\n", tststr );

				stv_unpack_int( msg->bufid, &cc, 1, 1 );
				printf( "\t# processors: %d\n", cc );

				stv_unpack_int( msg->bufid, &cc, 1, 1 );
				printf( "\t# epoch: %d\n\n", cc );
			}
		}
	}

	return( stvStatusOk );
}

